#!/usr/bin/perl

my @Id    = split / /,'$Id: nodelist2dns.pl,v 1.15 2013/02/10 18:21:16 stas Exp $';
my $ver   = $Id[2];
my $banner= "nodelist2dns v.$ver - A perl script to generate DNS distributed Nodelist zone file from fidonet nodelist";
# (c) Copyleft 2009-2013 Stas Degteff 2:5080/102@fidonet
# (c) Copyleft 2010 Robert James Clay 1:120/544@fidonet
# Distributes on GNU GPL v3, see http://www.gnu.org

# Features:
# - ignore "Down" and "Hold" nodes
# - make "A" and "CNAME" records like "fNNNN.nMMMM.zZ" for nodes Z:MMMM/NNNN
# - make "A" and "CNAME" records like "f0.nMMMM.zZ" for host nodes Z:MMMM/0
# - parse INA flag
# - IP address and domain name is taken from "system name" field of the nodelist
#   if not presented in the flags Ixx (optional, obsoleted)
# - IPv4 address is taken from "phone number" field of the nodelist if not presented
#   in the flags Ixx (optional, deprecated)

# ToDo:
# - parse flags "DOM", "DO4", "DO3", "DO2", "IRD"

# What it's works.
# Scans nodelist:
# - skip any nodes without flag "IBN", "IFC", "IFT", "ITN", "IVM";
# - for lines begins with "Zone,", "Region,", "Host,", "Hub,", "Pvt,", "," check ",INA:"
#   and match address after this flag;
# - if "INA" not presented and Ixx flag without host, then match IP (numberic) address
#   or fully-qualified domain name in field 3 ("station/BBS name") and match IPv4
#   address prefixed with "000-" in field 6 ("phone number").

# Changelog.
# 1.0 
# * first public release
# 1.1 
# + parse "INA" flag
# ! parse "Region" lines (bugfix)
# ! fixed ip address detection (regexp bugfix)
# + display warnings if incorrect nodelist line is found
# + if domain name or IP address isn't specified then generate default domain
#   name (using template fXXX.nYYY.zZ.fidonet.net. "fidonet.net" is specified
#   in configuration section and may be changed)
# + multiple addresses: first taken from "INA" flag, second taken from BBS name
#   field
# 1.2 
# ! Write CNAME only if A records not wroted for host.
# 1.3 
# + Add $TTL directive into zone file
# 1.4 
# + Skip empty (last) line of nodelist
# + Improve disgnostic
# 1.5 
# + Skip last line if one contains Ctrl-Z only
# 1.6
# + Debug mode
# ! Allow four-char TLD  (like "info")
# + Parse phone number string
# ! Fix domain name detection for station name field
# + Parse IBN flag
# 1.7
# + Use DNS::ZoneParse Perl Module to generate the new bind zone file
# + Pragmas - use warnings & use strict
#
# 1.8
# + Set site host for "A" record for domain self.
# - Default root domain changed to "binkp.net"
#
# 1.9
# + Configuration options is implemented using Getopt::Long
# + POD documentation and online help using Pod::Usage
# - Extraction hostname and IP address from 3rd field "BBS name" is maked optional,
# because too much false records and very few nodes with the valid hostname in
# the third field.
# - Default root domain usage maked optional for compatibility with FTS-5004
#
# 1.10
# ! Fix compilation warning
#
# 1.11
# ! Fix (change) generation serial of zone. Warning! It is need to remove old zone
# file on all secondaries.
# + Read old zone file and use serial from it if generate zone from same nodelist.
#
# 1.12
# - Don't use DNS::ZoneParse because this module don't supports SRV.
# + Parse port number in the flags.
# + Parse several same flags.
# + Parse flags INA,IBN,IFC,ITN,IVM,IFT.
# + Store several hostnames (create one CNAME and all SRV)
# + Parse IPv6 addresses
#
# 1.13
# ! Fix generation the SRV with default port from hostnames list 
# ! Fix generation the SRV for nodes without IBN flag
# + Option "--quiet" or "-q"
#
# 1.14
# ! Use SRV for nodes without IBN and with other Ixx flag (don't create  "A", "AAAA" and/or "CNAME" in this case)
#
# 1.15
# - Look in phone number are optional now.

use strict; #use warnings;
#use DNS::ZoneParse;
use Getopt::Long qw(:config no_ignore_case bundling);
use Pod::Usage;

# configuration variables
#
my $info="DDN nodelist.grumbler.org";    # zone's name (informational text line)
my $TTL    =7200;                       # TTL for the DNS zone SOA record
my $server ="srv.grumbler.org";         # DNS host
my $person ="hostmaster.grumbler.org";  # responsible person (email address)
my $minttl =3600;                       # minimum zone's TTL
my $refresh=86400;                      # zone's refresh period
my $retry  =3600;                       # zone's retry period
my $expire =604800;                     # zone's expire time
my $default_root_domain=undef;          # domain for constucts hostname if
                                        # IP address/domain isn't specified in
                                        # nodelist line.
my $bbs_name_field=undef;               # Flag for extract host name and IP from 3rd field (system name)
my $phone_number_field=undef;           # Flag for extract IPv4 from 6th field (phone number)
my $siteserver="212.49.98.146";         # IP address for DNS zone "A" record (web server usually)
my $help=undef;
my $debug=undef;
my $quiet=undef;
my %protocol=( IBN=>'binkp', IFC=>'ifcico', IFT=>'ftp', ITN=>'telnet', IVM=>'vmodem' );
my %defport=( IBN=>'24554', IFC=>'60179', IFT=>'21', ITN=>'23', IVM=>'3141' );
#
# Read options
my $optionsOK = GetOptions( "default-root-domain|R:s" => \$default_root_domain,
                            "BBS-name-field|3" => \$bbs_name_field,
                            "Phone-number-field|6" => \$phone_number_field,
                            "siteserver|s:s" => \$siteserver,
                            "help|H|?" => \$help,
                            "debug" => \$debug,
                            "quiet|q" => \$quiet,
                            "nameserver|ns|n:s" => \$server,
                            "info|I:s" => \$info,
                            "ttl|T:i" => \$TTL,
                            "person|p:s" => \$person,
                            "minttl|M:i" => \$minttl,
                            "refresh|R:i" => \$refresh,
                            "retry|r:i" => \$retry,
                            "expire|E:i" => \$expire,
                          );
#
# configuration end
#

if( $help ){
  pod2usage( -exitval => 1, -verbose => 2, -msg => $banner );
}
# check options
if( scalar($server) ){
  if( $server !~ /[[:alnum:].-]+\.[[:alpha:]]{2,}/ ){
    $optionsOK = 0;
    $banner .= "\n\nError: invalid hostname of nameserver.\n\n";
  }
} else {
  $optionsOK = 0;
  $banner .= "\n\nError: nameserver (hostname) should be specified\n\n";
}
if( scalar($server) ){
  $server =~ s/\.$//; # remove end point from server name
}
if( scalar($person) ){
  $person =~ s/@/./;  # if responsible person email specified like name@email.domain
  if( $person !~ /[[:alnum:].-]+\.[[:alpha:]]{2,}/ ){
    $optionsOK = 0;
    $banner .= "\n\nError: invalid email of responsible person.\n\n";
  }
} else {
  $optionsOK = 0;
  $banner .= "\n\nError: responsible person (email) should be specified\n\n";
}

if( ! $optionsOK || $#ARGV<0 || $#ARGV>=2){
  pod2usage( -exitval => 1, -verbose => 0, -msg => $banner );
}


my ($nodelist,$zonefile)=@ARGV;

if( -f $nodelist && -r $nodelist){
  open NODELIST, $nodelist || die "Can't open $nodelist: $!\n";
  print "Parse nodelist $nodelist\n" if(!$quiet);
}else{ die "$nodelist is not readable\n"; }

my $dns_file;
my $sn;
#if( -f $zonefile and -r $zonefile ){
#  $dns_file = DNS::ZoneParse->new( $zonefile );
#  $sn=$dns_file->soa();
#  $sn=${$sn}{'serial'};
#  undef $dns_file;
#}

open DNSZONE, ">$zonefile" || die "Can't create $zonefile: $!\n";

&process;

print "File '$zonefile' is generated\n" if(!$quiet);

close NODELIST;
close DNSZONE;
exit 0;

sub push_srv{
# 0 - link to @srv, 1 - FTNhost, 2 - hostname or IP, 3 - port, 4 - protocol, 5 - ('ipv4'|'ipv6'|.*)
  if( $_[5] eq 'ipv4' ){
    (my $h1 = $_[2]) =~ s/\./-/g;
    push @{$_[0]}, "$h1.$_[1]	IN	A	$_[2]";
    push @{$_[0]}, "_$_[4]._tcp.$_[1]	IN	SRV	0 0 $_[3] $h1.$_[1]";
  }elsif( $_[5] eq 'ipv6' ){
    (my $h1 = $_[2]) =~ s/:/-/g;
    push @{$_[0]}, "$h1.$_[1]	IN	AAAA	$_[2]";
    push @{$_[0]}, "_$_[4]._tcp.$_[1]	IN	SRV	0 0 $_[3] $h1.$_[1]";
  }else{
    my $h1 = ( $_[2]=~/\.$/ )? $_[2] : "$_[2]." ;
    push @{$_[0]}, "_$_[4]._tcp.$_[1]	IN	SRV	0 0 $_[3] $h1";
  }
}

sub process{

my $line = <NODELIST>;

# ;A FidoNet Nodelist for Friday, September 2, 2005 -- Day number 245 : 31474
# $sn="20052450";
if( $line =~ /;A FidoNet Nodelist for [\w\s,]+, (\d\d\d\d) -- Day number (\d+) /){
  my $year=$1;
  my $day=$2;
  die "Wrong nodelist: invalid year or day !\n" if( $year<=1980 or $day<=1 or $day>=366 );
  if( length($day)<3 ){
    $day="0".$day if($day<10);
    $day="0".$day if($day<100);
  }
  if( scalar($sn) and ( $year.$day eq substr($sn,0,7) ) ){
    $sn++;
  } else {
    $sn=$year.$day."0";
  }
}else{
  die "Wrong nodelist!\n";
}

$line =~ s/^;//;
$line =~ s/[\r\n]//g;

print DNSZONE "
; DDN compatible DNS zone file
; Generated by $banner
;
\$TTL $TTL
\@      $TTL    IN      SOA     $server. $person. (
                        $sn     ; serial number: concantenate of year and day of nodelist, and append a version digit
                        $refresh            ; refresh
                        $retry          ; retry
                        $expire         ; expire
                        $minttl         ; minimum TTL
                        )

\@              IN      NS      $server.
fidonet		IN	NS	$server.
\@              IN      TXT     \"This DDN generated using $banner\"
\@              IN      TXT     \"from $line\"

";

if( $siteserver =~ /^[1-9][0-9]*\.[0-9]+\.[0-9]+\.[0-9]+$/ ){
 print DNSZONE "\@		IN	A	$siteserver\n";
}

  my ($zone,$net,$node);
  my $linecount=1;
  while(<NODELIST>){
    $linecount++;
    next if(/^;/);
    next if( /^[\r\n\x1a]*$/ and eof(NODELIST) );
    if(/^[^\w,]/) { print "Warning! Illegal line ${linecount}:\n", $_;};

    print STDERR "Line \"$_\"\n" if( $debug );
    if(/^Zone,(\d+),/){
      $zone=$1;
      $net=$1;
      $node=0;
      print STDERR "Zone $zone\n" if( $debug );
    }
    elsif(/^(Region|Host),(\d+),/){ # net/independent number
      $net=$2;
      $node=0;
      print STDERR "Net(Region/Host) $zone:$net\n" if( $debug );
    }
    elsif(/^(Down|Hold),/){      # skip nodes with Down/Hold status.
      print STDERR "Down/Hold,skip\n" if( $debug );
      next
    }
    elsif(/^(Hub|Pvt|),(\d+),/){ # node number
      $node=$2;
      print STDERR "Node: $zone:$net/$node\n" if( $debug );
    }
    else{
      print STDERR "...skip\n" if( $debug );
      next;
    }
    if(!/,I[[:alpha:]]{2}(:.*)?([,\r\n].*)?$/){ # skip nodes without internet flags
      print STDERR 'Node without "I??" flag, skip\n' if( $debug );
      next
    }

    s/[\r\n]*$//g;
    my @fields = split ',';
    my $host = "f$node.n$net.z$zone";
    my @hostnames;
    my @ipv4;
    my @ipv6;
    my @srv;

    my %ina;
    foreach my $f (@fields[7..$#fields]){
      if( $f =~ /^INA:([^:]+|\[[[:xdigit:]:]+\])(:(\d+))?$/ ){
        my $h=$1;
        my $p;
        if( scalar($3) ){
          $p=$3;
        }
        if( $h =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/){
          $ina{"$h:$p"}='ipv4';
        }elsif( $h =~ /^[[:alnum:]][[:alnum:].-]+\.[[:alpha:]]{2,}$/){
          $ina{"$h:$p"}='hostname';
        }elsif( $h =~ /^\[([[:xdigit:]]{1,2}:[[:xdigit:]:]+)\]$/){
          $ina{"$1:$p"}='ipv6';
        }
      }
    }
    if( $bbs_name_field and !defined(%ina) ){
      if( $fields[2]=~/^([[:alnum:]][[:alnum:]\.-]+\.[[:alpha:]]{2,6})$/ ){
         $ina{lc($1).":"}='hostname';
      }elsif( $fields[2]=~/^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})$/ ){
         $ina{"$1:"}='ipv4';
      }elsif( $fields[2]=~/^\[([[:xdigit:]]{1,2}:[[:xdigit:]:]+)\]$/ ){
         $ina{lc($1).":"}='ipv6';
      }
      print STDERR " From station name field: $1\n" if( $debug );
    }
    if( $phone_number_field and !defined(%ina) and $fields[5]=~/^000-([0-2]\d{1,2})-([0-2]\d{1,2})-([0-2]\d{1,2})-([0-2]\d{1,2})$/ ){
      # Phone number field
      if( $1>0 && $1<255 && $2>=0 && $2<=255 && $3>=0 && $3<=255 && $4>0 && $4<255 ){
         $ina{"$1.$2.$3.$4:"}='ipv4';
        print STDERR " From phone number field: $1.$2.$3.$4\n" if( $debug );
      }
    }
    foreach my $f (@fields[7..$#fields]){
      if( $f =~ /^(ITN|IBN|IFC|IVM|IFT)(:([^:]+\.[^:]*|\[[[:xdigit:]:]+\]))?(:(\d+))?$/ ){
        my $flag=$1;
        my $h;
        my $p;
        if( scalar($3) ){
          $h=$3;
        }
        if( scalar($5) ){
          $p=$5;
        }
        if( $h =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/){
          # flag with IPv4
          if( !scalar($p) and $flag eq 'IBN' ){
            push @ipv4, $h;
            print STDERR " A from $f\n" if( $debug );
          }else{
            push_srv \@srv, $host, $h, (scalar($p)?$p:$defport{$flag}), $protocol{$flag}, 'ipv4' ;
            print STDERR " A & SRV from $f\n" if( $debug );
          }
        }elsif( $h =~ /^[[:alnum:]][[:alnum:].-]+\.[[:alpha:]]{2,}$/){
          # flag with hostname
          if( !scalar($p) and $flag eq 'IBN' ){
            push @hostnames, lc($h);
            print STDERR " hostname $h from $f\n" if( $debug );
          }else{
            push_srv \@srv, $host, $h, scalar($p)?$p:$defport{$flag}, $protocol{$flag} ;
            print STDERR " SRV from $f\n" if( $debug );
          }
        }elsif( $h =~ /^\[([[:xdigit:]]{1,2}:[[:xdigit:]:]+)\]$/){
          # flag with IPv6
          if( !scalar($p) and $flag eq 'IBN' ){
            push @ipv6, $1;
            print STDERR " AAAA from $f\n" if( $debug );
          }else{
            push_srv \@srv, $host, $1, scalar($p)?$p:$defport{$flag}, $protocol{$flag}, 'ipv6' ;
            print STDERR " AAAA & SRV from $f\n" if( $debug );
          }
        }elsif( !scalar($h) ){
          foreach my $hp (keys(%ina)){
            (my $inah, my $inap) = split ":", $hp;
            if( $flag eq 'IBN' and !scalar($p) and !scalar($inap) ){
               if( $ina{$hp} eq 'ipv4' ){
                 push @ipv4, $inah;
               }elsif( $ina{$hp} eq 'ipv6' ){
                 push @ipv6, $inah;
               }else{
                 push @hostnames, lc($inah);
               }
            }else{
               push_srv \@srv, $host, $inah, (scalar($p))?$p:(scalar($inap)?$inap:$defport{$flag}), $protocol{$flag}, $ina{$hp};
            } 
          }
        }else{
           print "Wrong hostname is detected: $h ($_)\n";
        }        
      }
    }
    if( $#srv>-1 and ($#hostnames>-1 or $#ipv6>-1 or $#ipv4>-1) ){ # add SRV if exist IBN without port and another SRV created
      foreach my $i (@ipv4){
        push_srv \@srv, $host, $i, $defport{'IBN'}, $protocol{'IBN'}, 'ipv4';
      }
      foreach my $i (@ipv6){
        push_srv \@srv, $host, $i, $defport{'IBN'}, $protocol{'IBN'}, 'ipv6';
      }
      if( $#hostnames>-1 and $#ipv6==-1 and $#ipv4==-1 ){
        foreach my $i (@hostnames){
          push_srv \@srv, $host, $i, $defport{'IBN'}, $protocol{'IBN'};
        }
      }
    }

    if($#hostnames==-1 && scalar($default_root_domain) && $default_root_domain =~ /^[[:alnum:].-]+$/ ){
        push @hostnames, "f$node.n$net.z$zone.$default_root_domain.";
      print STDERR " Use default from $default_root_domain.\n" if( $debug );
    }

    { # output block
      my $h;
      my @ip = sort(@ipv4);
      while( $h = shift(@ip) ) {
        if( $h ne $ip[0] ){
          print DNSZONE "$host 	IN 	A 	$h\n";
          print STDERR " Use $h in A\n" if( $debug );
        }else{ print STDERR " Duplicated $h\n" if( $debug ); }
      }
      @ip = sort(@ipv6);
      while( $h = shift(@ip) ) {
        if( $h ne $ip[0] ){
          print DNSZONE "$host 	IN 	AAAA 	$h\n";
          print STDERR " Use $h in AAAA\n" if( $debug );
        }else{ print STDERR " Duplicated $h\n" if( $debug ); }
      }
      if( $#ipv4==-1 and $#ipv6==-1 and $#hostnames>-1 ){
        $h = $hostnames[0];
        $h =~ s/\.$//;
        print DNSZONE "$host 	IN 	CNAME 	$h.\n";
        print STDERR " Use $h in CNAME\n" if( $debug );
      }
      if( $#hostnames>-1 and ($#hostnames>0 or $#ipv4>-1 or $#ipv6>-1) ) {
        @ip = sort(@hostnames);
        while( $h = shift(@ip) ) {
          next if( $h eq $ip[0] );
          $h =~ s/\.$//;
          push @srv, "_$protocol{'IBN'}._tcp.$host	IN	SRV	0 0 $defport{'IBN'} $h.";
          print STDERR " Use $h in SRV binkp default\n" if( $debug );
        }
      }
      @ip = sort(@srv);
      while( $h = shift(@ip) ) {
        next if( $h eq $ip[0] );
        print DNSZONE $h, "\n";
        print STDERR " Write $h\n" if( $debug );
      }
    }
  }
}

__END__

=head1 NAME

nodelist2dns.pl - a script to generate DNS distributed Nodelist zone from (fidonet) nodelist

=head1 SYNOPSIS

nodelist2dns.pl {--help|-H}

nodelist2dns.pl [options] nodelist.file dns.zone.file

=head1 DESCRIPTION

nodelist2dns.pl reads lines from the nodelist file (1st non-option parameter), analize this line and, if TCP flags for node is exist, creates record(s) for the node in the DNS zone file (2nd non-option parameter). The serial number of the zone generated from nodelist data: year and day number. Old DNS zone file overwrites.

=head1 OPTIONS

=over 4

=item  B<--BBS-name-field>

=item  B<-3>

Extract hostname or IP address from 3rd field of node line (BBS name field)

This way is obsoleted because very few nodes write to 3rd field real hostname or IP and
3rd field of nodelist contain too much false hosts or contain host not for mailer connection.

=item  B<--debug>

Enable debug output (to stderr).

=item  B<--default-root-domain="domain.tld">

=item  B<-R "domain.tld">

Set the default root domain - if node don't specify hostname or IP, then this domain used for consruct hostname, like f0.n50.z2.root-domain.
Optional.

=item  B<--expire=expiration_timeout>

=item  B<-E expiration_timeout>

Zone expiration timeout for SOA record, in seconds (pozitive integer).
Default value: 604800.

=item  B<--help>

=item  B<-H>

Print this information

=item  B<--info="info string">

=item  B<-I "info string">

Informational string fot the title of the zone file and for the "TXT" record for zone.
Optional.

=item  B<--minttl=minimum-TTL>

=item  B<-M minimum-TTL>

Minimum TTL of the zone for SOA record, in seconds (pozitive integer).
Default value: 3600.

=item  B<--nameserver>

=item  B<--ns>

=item  B<-n>

Hostname of the nameserver for generated domain zone.
Mandatory.

=item  B<-Phone-number-field>

=item  B<-6>

Extract IPv4 address from 6th field of node line (Phone number field)

This way is deprecated because in the many countries special phone calls begins from zero.

=item  B<--debug>


=item  B<--person="responsible person">

=item  B<-p "responsible person">

Email of the responsible person of the domain zone (hostmaster's email).
Mandatory.

=item  B<--quiet>
=item  B<-q>

Don't display messages, only errors.
Optional.

=item  B<--refresh=refresh_period>

=item  B<-R refresh_period>

Refresh period for SOA record, in seconds (pozitive integer).
Default value: 86400.

=item  B<--retry=retry_interval>

=item  B<-r retry_interval>

Zone retry interval for SOA record, in seconds (pozitive integer).
Default value: 3600.

=item  B<--siteserver=web_server_address>

=item  B<-s web_server_address>

IP address of the web server. If specified, an "A" record for domain zone will be generated. Hostname is not allowed because CNAME can't be mixed with any other records (SOA, NS).
Optional.

=item  B<--ttl=default-TTL>

=item  B<-T default-TTL>

Default TTL of the zone for SOA record, in seconds (pozitive integer).
Default value: 7200.

=back

=head1 AUTHOR

 (c) Copyleft 2009-2013 Stas Degteff 2:5080/102@fidonet
 (c) Copyleft 2010 Robert James Clay 1:120/544@fidonet

=head1 LICENSE

 GPLv3 (http://www.gnu.org/licenses/gpl.txt)

=cut
